new;
library pf,pgraph;

rndseed 123;

declare matris USE_BOUNDS;

loadm data = example5-data;
loadm theta = example5-theta;

R_Fund = data[.,1];
R_MSCI = data[.,2 3];
True_Coeffs = data[.,4 5];
KF_Coeffs = data[.,6 7];

nDates = rows(Data);
t = seqa(1,1,nDates);

a0 = 0.5|0.5;
P0 = 0*ones(2,2);

nFactors = cols(R_MSCI);

Ht_KF = theta[1]^2;

P = theta[2:4];
P = lowmat(xpnd(P));
Qt_KF = P*P';

/***********************************************/
/*                PARTICLE FILTER              */
/***********************************************/

/* 
**> Sample the importance density
**  Partial Gaussian case permits to choose the optimal importance density (p(x_]k[|y_]k[,x_]k-1[)
**  instead of the usual prior
*/

proc (1) =  pf_importance_sampling(t,x,z);

  if _pf_Partial_Gaussian == 0;
    retp(pf_prior_sampling(t,x));
  endif;

  local Ft,Qt,Rt,Ht,invQt,invSigma,Sigma,mu,nS,i;
  local F,Q,R,H;
  F = _pf_ft;
  Q = _pf_Qt;
  R = _pf_Rt;
  H = _pf_Ht;
  local F:proc,Q:proc,R:proc,H:proc;
  
  Ft = F(t,x);
  Qt = Q(t,x);
  Rt = R(t,x);
  Ht = H(t,error(0));
  
  invQt = invpd(Qt);
  invSigma = invQt + Ht' * 1/Rt * Ht;
  Sigma = inv(invSigma);
  mu = invQt*Ft' + Ht'*(1/Rt)*z;
  mu = Sigma * mu;

  nS = rows(x);  
  i = 1;
  do until i > nS;   
    x[i,.] = rndmvn(mu[.,i],sigma,1);
    i = i + 1;
  endo;

  retp(x);
endp;

/* 
**> Sample the prior density
*/

proc (1) =  pf_prior_sampling(t,x);
  local nS,Ft,Qt;
  local cnd,nOut,indx,mu;
  local F,Q;
  F = _pf_ft;
  Q = _pf_Qt;
  local F:proc, Q:proc;

  nS = rows(x);
  
  Ft = F(t,x);
  Qt = Q(t,x);  

  x = rndmvn(Ft',Qt,nS);

  if USE_BOUNDS == 1;
    cnd = (x[.,1] .>= 0) .and (x[.,1] .<= 1) .and (x[.,2] .>= 0) .and (x[.,2] .<= 1) ;
    nOut = nS - sumc(cnd);
    do while nOut > 0;
      indx = selif(seqa(1,1,nS),1 - cnd);
      mu = Ft[indx,.]';
      x[indx,.] = rndmvn(mu,Qt,nOut);
      cnd = (x[.,1] .>= 0) .and (x[.,1] .<= 1) .and (x[.,2] .>= 0) .and (x[.,2] .<= 1) ;
      nOut = nS - sumc(cnd);
    endo;
  endif;

  retp(x);
endp;

/* 
**> Evaluate the importance density function: q(x_{t}|x{t-1},z_{t})
*/

proc (1) = pf_importance_density(t,nx,x,z);
  local pdf;
  pdf = pf_prior_density(t,nx,x);
  retp(pdf);
endp;

/* 
**> Evaluate the prior density function: p(x_{t}|x{t-1})
**
** Remark: It is the same as the importance density function
*/

proc (1) = pf_prior_density(t,nx,x);
  local Ft,Qt,pdf;
  local nS,cnd,i,cdf_UR,cdf_UL,cdf_DR,cdf_DL;
  local F,Q;
  F = _pf_ft;
  Q = _pf_Qt;
  local F:proc,Q:proc;

  Ft = F(t,x);
  Qt = Q(t,x);

  pdf = pdfmvn(nx,x,Qt);

  if USE_BOUNDS == 1;
    cnd = (nx[.,1] .>= 0) .and (nx[.,1] .<= 1) .and (nx[.,2] .>= 0) .and (nx[.,2] .<= 1) ;
    nS = rows(pdf);
    i = 1;
    do until i > nS;
      pdf[i] = pdf[i] .* cnd[i];
      cdf_UR = cdfmvn2(1|1,x[i,.]',Qt);
      cdf_UL = cdfmvn2(0|1,x[i,.]',Qt);
      cdf_DR = cdfmvn2(1|0,x[i,.]',Qt);
      cdf_DL = cdfmvn2(0|0,x[i,.]',Qt);
      pdf[i] = pdf[i] / (cdf_UR + cdf_DL - cdf_UL-cdf_DR);
      i = i + 1;
    endo;
  endif;
  
  retp(pdf);
endp;

/* 
**> Evaluate the likelihood function p(z_{t}|x_{t})
*/

proc (1) = pf_likelihood_density(t,nx,z);
  local Ht,Rt,pdf;
  local H,R;
  H = _pf_ht;
  R = _pf_Rt;
  local H:proc,R:proc;
  
  Ht = H(t,nx);
  Rt = R(t,nx);
  
  pdf = pdfn2(z,Ht,sqrt(Rt));

  retp(pdf);
endp;

/* 
**> Sample the initial state vector
*/

proc (1) = pf_initial_sampling(nS);
  local x;
   x = .5*ones(nS,2);
  retp(x);
endp;

/*
**> State Space Model
*/

fn pf_ft(t,x) = x;
fn pf_Qt(t,x) = Qt_KF;
fn pf_Rt(t,x) = Ht_KF;

proc (1) = pf_ht(t,x);
  if x == error(0);
    retp(R_MSCI[t,.]);
  else;
    retp(x * R_MSCI[t,.]');
  endif;
endp;

/*
**> Main Test
**
*/

_pf_Partial_Gaussian = 1;

USE_Bounds = 1;

if USE_Bounds;
  _rpf_Bounds = (0~1) | 
                (0~1);
endif;

call Particle_Filter_Set(&pf_Ft,&pf_Qt,&pf_Ht,&pf_Rt,
                        &pf_prior_density,&pf_prior_density,&pf_likelihood_density,0,
                        &pf_importance_sampling,&pf_prior_sampling,&pf_initial_sampling);

cls;

x0 = pf_initial_sampling(1);
//nT = rows(Coeffs);
nS = 1000;

{SIS_x,SIS_w,SIS_Coeffs,SIS_cov} = SIS_Particle_Filter(R_Fund,nS);                /* Running the SIS particle filter */ 
{PF_x,PF_w,PF_Coeffs,PF_cov} = Generic_Particle_Filter(R_Fund,nS);                /* Running the generalized particle filter */   
{SIR_x,SIR_w,SIR_Coeffs,SIR_cov} = SIR_Particle_Filter(R_Fund,nS);                /* Running the SIR particle filter */     
{RPF_x,RPF_w,RPF_Coeffs,RPF_cov} = Regularized_Particle_Filter(R_Fund,nS);        /* Running the regularised particle filter */   

graphset;
   window(2,2,0);
  _pdate = 0; _pframe = 0; _pnum = 2;
  _plwidth = 10; _pltype = 6|3*ones(5,1); 
  _ptitlht = 0.30; _paxht = 0.225; _pnumht = 0.225;
  xtics(1,2*12*10,12*2,1);
  let xlab = "90" "92" "94" "96" "98" "00" 
             "02" "04" "06" "08";
  asclabel(xlab,0);
  ytics(0,100,20,2);

  setwind(1);
   title("\216SIS Algorithm");
  _pcolor = 1|2|3;
  xy(0,100*(True_Coeffs[.,1]~KF_Coeffs[.,1]~SIS_Coeffs[.,1]));

  setwind(2);
   title("\216GPF Algorithm");
  _pcolor = 1|2|4;
  xy(0,100*(True_Coeffs[.,1]~KF_Coeffs[.,1]~PF_Coeffs[.,1]));

  setwind(3);
   title("\216SIR Algorithm");
  _pcolor = 1|2|5;
  xy(0,100*(True_Coeffs[.,1]~KF_Coeffs[.,1]~SIR_Coeffs[.,1]));

  setwind(4);
   title("\216RPF Algorithm");
  _pcolor = 1|2|6;
  _plegctl = {2 6 7 1.0};
  _plegstr = "True\000KF\000PF";
  xy(0,100*(True_Coeffs[.,1]~KF_Coeffs[.,1]~RPF_Coeffs[.,1]));

  graphprt("-c=1 -cf=example5.eps");

endwind;